# pysomr - Python Wrapper for SoMR API
# Copyright (C) 2026  black-sliver
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# cython: language_level=3
# distutils: language = c
# distutils: depends = SecretOfManaRandomizer/SoMRandomizer.api/c/api.h

# NOTE: cython support in IDEs is not great, specifically tracking cdef types,
#       so we add a bunch of "no inspection" markers everywhere

import pathlib
import typing as t

import cython
from libc.stdint cimport uint_least16_t, int32_t, uint8_t


ctypedef uint_least16_t char16_t
# """Strings are "windows wide char*" that is pointers to 16bit characters encoded in "Windows Unicode" (UCS2/WTF16)"""


cdef extern from *:
    """
    // avoid warning from cython-generated code with MSVC + pyximport
    #ifdef _MSC_VER
    #pragma warning( disable: 4551 )
    #endif
    """


cdef extern from "SecretOfManaRandomizer/SoMRandomizer.api/c/api.h":
    ctypedef struct SoMR_ItemList:
        pass
    ctypedef struct SoMR_Item:
        pass
    ctypedef struct SoMR_LocationList:
        pass
    ctypedef struct SoMR_Location:
        pass
    ctypedef struct SoMR_OWSettings:
        pass
    ctypedef struct SoMR_OWGenerator:
        pass
    ctypedef struct SoMR_Context:
        pass
    ctypedef struct SoMR_WorkingData:
        pass
    ctypedef struct SoMR_StrList:
        pass

    # noinspection PyPep8Naming
    SoMR_ItemList* SoMR_OW_GetAllItems()
    # noinspection PyPep8Naming
    SoMR_LocationList* SoMR_OW_GetAllLocations()
    # noinspection PyPep8Naming
    SoMR_OWSettings* SoMR_OW_NewSettings()
    # noinspection PyPep8Naming
    SoMR_OWGenerator* SoMR_OW_NewGenerator()
    # noinspection PyPep8Naming
    SoMR_Context* SoMR_OW_Init(char16_t* src, char16_t* seed, SoMR_OWGenerator* generator, SoMR_OWSettings* settings)
    # noinspection PyPep8Naming
    void SoMR_OW_Run(
            char16_t* src,
            char16_t* seed,
            SoMR_OWGenerator* generator,
            SoMR_OWSettings* settings,
            SoMR_Context* context,
    )

    # noinspection PyPep8Naming
    int32_t SoMR_ItemList_Count(SoMR_ItemList* lst)
    # noinspection PyPep8Naming
    SoMR_Item* SoMR_ItemList_At(SoMR_ItemList* lst, int32_t index)
    # noinspection PyPep8Naming
    void SoMR_ItemList_Unref(SoMR_ItemList* lst)
    # noinspection PyPep8Naming
    char16_t* SoMR_Item_GetName(SoMR_Item* item)
    # noinspection PyPep8Naming
    char16_t * SoMR_Item_GetType(SoMR_Item* item)
    # noinspection PyPep8Naming
    uint8_t SoMR_Item_GetEventFlag(SoMR_Item* item)
    # noinspection PyPep8Naming
    uint8_t SoMR_Item_GetItemId(SoMR_Item* item)
    # noinspection PyPep8Naming
    void SoMR_Item_Unref(SoMR_Item* item)

    # noinspection PyPep8Naming
    int32_t SoMR_LocationList_Count(SoMR_LocationList* lst)
    # noinspection PyPep8Naming
    SoMR_Location* SoMR_LocationList_At(SoMR_LocationList* lst, int32_t index)
    # noinspection PyPep8Naming
    void SoMR_LocationList_Unref(SoMR_LocationList* lst)
    # noinspection PyPep8Naming
    char16_t* SoMR_Location_GetName(SoMR_Location* location)
    # noinspection PyPep8Naming
    int32_t SoMR_Location_GetMapNum(SoMR_Location * location)
    # noinspection PyPep8Naming
    int32_t SoMR_Location_GetObjNum(SoMR_Location * location)
    # noinspection PyPep8Naming
    int32_t SoMR_Location_GetEventNum(SoMR_Location* location)
    # noinspection PyPep8Naming
    int32_t SoMR_Location_GetEventIndex(SoMR_Location* location)
    # noinspection PyPep8Naming
    uint8_t SoMR_Location_GetLocationId(SoMR_Location* location)
    # noinspection PyPep8Naming
    void SoMR_Location_Unref(SoMR_Location* location)
    # noinspection PyPep8Naming
    SoMR_StrList* SoMR_Location_GetRequirements(SoMR_Location* location)

    # noinspection PyPep8Naming
    void SoMR_OWSettings_SetStr(SoMR_OWSettings* settings, char16_t* key, char16_t* value)
    # noinspection PyPep8Naming
    char16_t* SoMR_OWSettings_Dump(SoMR_OWSettings* settings)
    # noinspection PyPep8Naming
    void SoMR_OWSettings_Unref(SoMR_OWSettings* settings)

    # noinspection PyPep8Naming
    SoMR_ItemList* SoMR_OWGenerator_GetItems(SoMR_OWGenerator* generator)
    # noinspection PyPep8Naming
    SoMR_LocationList* SoMR_OWGenerator_GetLocations(SoMR_OWGenerator* generator)
    # noinspection PyPep8Naming
    void SoMR_OWGenerator_Unref(SoMR_OWGenerator* generator)

    # noinspection PyPep8Naming
    char16_t * SoMR_Context_GetError(SoMR_Context* context)
    # noinspection PyPep8Naming
    SoMR_WorkingData* SoMR_Context_GetWorkingData(SoMR_Context* context)
    # noinspection PyPep8Naming
    void SoMR_Context_Unref(SoMR_Context* context)

    # noinspection PyPep8Naming
    char16_t* SoMR_WorkingData_GetStr(SoMR_WorkingData* data, char16_t* key)
    # noinspection PyPep8Naming
    void SoMR_WorkingData_SetStr(SoMR_WorkingData * data, char16_t * key, char16_t * value)
    # noinspection PyPep8Naming
    char16_t* SoMR_WorkingData_Dump(SoMR_WorkingData* data)
    # noinspection PyPep8Naming
    void SoMR_WorkingData_Unref(SoMR_WorkingData* data)

    # noinspection PyPep8Naming
    int32_t SoMR_StrList_Count(SoMR_StrList* lst)
    # noinspection PyPep8Naming
    char16_t* SoMR_StrList_At(SoMR_StrList* lst, int32_t index)
    # noinspection PyPep8Naming
    void SoMR_StrList_Unref(SoMR_StrList* lst)

    # noinspection PyPep8Naming
    void SoMR_Str_Free(char16_t* string)


# TODO: move into enum?
WORKING_DATA_KEY_BOY_CLASS = "boyClass"
WORKING_DATA_KEY_GIRL_CLASS = "girlClass"
WORKING_DATA_KEY_SPRITE_CLASS = "spriteClass"
WORKING_DATA_KEY_BOY_EXISTS = "boyExists"  # exists + !in_logic = start_with
WORKING_DATA_KEY_GIRL_EXISTS = "girlExists"
WORKING_DATA_KEY_SPRITE_EXISTS = "spriteExists"
WORKING_DATA_KEY_BOY_IN_LOGIC = "findBoy"
WORKING_DATA_KEY_GIRL_IN_LOGIC = "findGirl"
WORKING_DATA_KEY_SPRITE_IN_LOGIC = "findSprite"
WORKING_DATA_KEY_BOY_START_WEAPON_INDEX = "boyStartWeapon"
WORKING_DATA_KEY_GIRL_START_WEAPON_INDEX = "girlStartWeapon"
WORKING_DATA_KEY_SPRITE_START_WEAPON_INDEX = "spriteStartWeapon"
# possible class values: OGboy, OGgirl, OGsprite, random, randomunique
OW_SETTINGS_KEY_BOY_CLASS = "opBoyRole"
OW_SETTINGS_KEY_GIRL_CLASS = "opGirlRole"
OW_SETTINGS_KEY_SPRITE_CLASS = "opSpriteRole"


cdef size_t ucs2_len(char16_t* data):
    cdef size_t i = 0
    while data[i] != 0:
        i += 1
    return i


cdef str ucs2_to_str(char16_t* data):
    return (<char *> data)[:ucs2_len(data) * 2].decode("UTF-16")


# noinspection PyUnresolvedReferences
cdef bytes str_to_ucs2(str data):
    # FIXME: there has to be a better way to do this
    cdef bytes encoded
    encoded = data.encode("UTF-16")
    if encoded[:2] == b"\xff\xfe" or encoded[:2] == b"\xfe\xff":
        return encoded[2:] + b'\0'  # strip BOM
    return encoded + b'\0'


@cython.auto_pickle(False)
cdef class Item:
    cdef SoMR_Item* _handle

    @staticmethod
    cdef Item from_handle(SoMR_Item* handle):
        self = Item()
        self._handle = handle
        return self

    def __del__(self) -> None:
        SoMR_Item_Unref(self._handle)

    @property
    def name(self) -> str:
        name_ucs2 = SoMR_Item_GetName(self._handle)
        try:
            # noinspection PyTypeChecker
            return ucs2_to_str(name_ucs2)
        finally:
            SoMR_Str_Free(name_ucs2)

    @property
    def type(self) -> str:
        type_ucs2 = SoMR_Item_GetType(self._handle)
        try:
            # noinspection PyTypeChecker
            return ucs2_to_str(type_ucs2)
        finally:
            SoMR_Str_Free(type_ucs2)

    @property
    def event_flag(self) -> int:
        return SoMR_Item_GetEventFlag(self._handle)

    @property
    def id(self) -> int:
        return SoMR_Item_GetItemId(self._handle)


@cython.auto_pickle(False)
cdef class Location:
    cdef SoMR_Location* _handle

    @staticmethod
    cdef Location from_handle(SoMR_Location* handle):
        self = Location()
        self._handle = handle
        return self

    def __del__(self) -> None:
        SoMR_Location_Unref(self._handle)

    @property
    def name(self) -> str:
        name_ucs2 = SoMR_Location_GetName(self._handle)
        try:
            # noinspection PyTypeChecker
            return ucs2_to_str(name_ucs2)
        finally:
            SoMR_Str_Free(name_ucs2)

    @property
    def map_num(self) -> int:
        return SoMR_Location_GetMapNum(self._handle)

    @property
    def obj_num(self) -> int:
        return SoMR_Location_GetObjNum(self._handle)

    @property
    def event_num(self) -> int:
        return SoMR_Location_GetEventNum(self._handle)

    @property
    def event_index(self) -> int:
        return SoMR_Location_GetEventIndex(self._handle)

    @property
    def id(self) -> int:
        return SoMR_Location_GetLocationId(self._handle)

    @property
    def requirements(self) -> StrList:
        # noinspection PyTypeChecker
        return StrList.from_handle(SoMR_Location_GetRequirements(self._handle))


@cython.auto_pickle(False)
cdef class ItemList:
    cdef SoMR_ItemList* _handle

    @staticmethod
    cdef ItemList from_handle(SoMR_ItemList* handle):
        self = ItemList()
        self._handle = handle
        return self

    def __del__(self) -> None:
        SoMR_ItemList_Unref(self._handle)

    def __len__(self) -> int:
        return SoMR_ItemList_Count(self._handle)

    def __iter__(self) -> t.Iterator[Item]:
        for i in range(0, len(self)):
            yield Item.from_handle(SoMR_ItemList_At(self._handle, i))  # noqa


@cython.auto_pickle(False)
cdef class LocationList:
    cdef SoMR_LocationList* _handle

    @staticmethod
    cdef LocationList from_handle(SoMR_LocationList* handle):
        self = LocationList()
        self._handle = handle
        return self

    def __del__(self) -> None:
        SoMR_LocationList_Unref(self._handle)

    def __len__(self) -> int:
        return SoMR_LocationList_Count(self._handle)

    def __iter__(self) -> t.Iterator[Location]:
        for i in range(0, len(self)):
            yield Location.from_handle(SoMR_LocationList_At(self._handle, i))  # noqa


@cython.auto_pickle(False)
cdef class StrList:
    cdef SoMR_StrList* _handle

    @staticmethod
    cdef StrList from_handle(SoMR_StrList* handle):
        self = StrList()
        self._handle = handle
        return self

    def __del__(self) -> None:
        SoMR_StrList_Unref(self._handle)

    def __len__(self) -> int:
        return SoMR_StrList_Count(self._handle)

    def __iter__(self) -> t.Iterator[str]:
        for i in range(0, len(self)):  # TODO: while not NULL instead
            value_ucs2 = SoMR_StrList_At(self._handle, i)
            try:
                # noinspection PyTypeChecker
                yield ucs2_to_str(value_ucs2)
            finally:
                SoMR_Str_Free(value_ucs2)


@cython.auto_pickle(False)
cdef class OWSettings:
    cdef SoMR_OWSettings* _handle

    @staticmethod
    cdef OWSettings from_handle(SoMR_OWSettings* handle):
        self = OWSettings()
        self._handle = handle
        return self

    def __del__(self) -> None:
        SoMR_OWSettings_Unref(self._handle)

    def __str__(self) -> str:
        s_ucs2 = SoMR_OWSettings_Dump(self._handle)
        try:
            # noinspection PyTypeChecker
            return ucs2_to_str(s_ucs2)
        finally:
            SoMR_Str_Free(s_ucs2)

    cdef SoMR_OWSettings* get_handle(self):
        return self._handle

    def __setitem__(self, key: str, value: str) -> None:
        # noinspection PyTypeChecker
        key_bytes = str_to_ucs2(key)
        cdef unsigned char* key_ptr = key_bytes
        # noinspection PyTypeChecker
        value_bytes = str_to_ucs2(value)
        cdef unsigned char* value_ptr = value_bytes
        SoMR_OWSettings_SetStr(self._handle, <char16_t*>key_ptr, <char16_t*>value_ptr)


@cython.auto_pickle(False)
cdef class OWGenerator:
    cdef SoMR_OWGenerator* _handle

    @staticmethod
    cdef OWGenerator from_handle(SoMR_OWGenerator* handle):
        self = OWGenerator()
        self._handle = handle
        return self

    def __del__(self) -> None:
        SoMR_OWGenerator_Unref(self._handle)

    cdef SoMR_OWGenerator* get_handle(self):
        return self._handle

    def get_items(self) -> ItemList:
        # noinspection PyTypeChecker
        return ItemList.from_handle(SoMR_OWGenerator_GetItems(self._handle))

    def get_locations(self) -> LocationList:
        # noinspection PyTypeChecker
        return LocationList.from_handle(SoMR_OWGenerator_GetLocations(self._handle))


@cython.auto_pickle(False)
cdef class Context:
    cdef SoMR_Context* _handle
    _working_data: WorkingData | None

    @staticmethod
    cdef Context from_handle(SoMR_Context* handle):
        self = Context()
        self._handle = handle
        return self

    def __init__(self):
        self._working_data = None

    def __del__(self) -> None:
        SoMR_Context_Unref(self._handle)

    cdef SoMR_Context* get_handle(self):
        return self._handle

    @property
    def working_data(self) -> WorkingData:
        if self._working_data is None:
            # noinspection PyTypeChecker
            self._working_data = WorkingData.from_handle(SoMR_Context_GetWorkingData(self._handle))
        return self._working_data

    @property
    def error(self) -> str | None:
        msg_ucs2 = SoMR_Context_GetError(self._handle)
        if not msg_ucs2:
            return None
        try:
            # noinspection PyTypeChecker
            return ucs2_to_str(msg_ucs2)
        finally:
            SoMR_Str_Free(msg_ucs2)


@cython.auto_pickle(False)
cdef class WorkingData:
    cdef SoMR_WorkingData* _handle

    @staticmethod
    cdef WorkingData from_handle(SoMR_WorkingData* handle):
        self = WorkingData()
        self._handle = handle
        return self

    def __del__(self) -> None:
        SoMR_WorkingData_Unref(self._handle)

    def __str__(self) -> str:
        s_ucs2 = SoMR_WorkingData_Dump(self._handle)
        try:
            # noinspection PyTypeChecker
            return ucs2_to_str(s_ucs2)
        finally:
            SoMR_Str_Free(s_ucs2)

    def __getitem__(self, item: str) -> str:
        # TODO: make keys opaque and cache in .net land?
        # noinspection PyTypeChecker
        key_bytes = str_to_ucs2(item)
        cdef unsigned char* key_ptr = key_bytes
        val_ucs2 = SoMR_WorkingData_GetStr(self._handle, <char16_t*>key_ptr)
        try:
            # noinspection PyTypeChecker
            return ucs2_to_str(val_ucs2)
        finally:
            SoMR_Str_Free(val_ucs2)

    def __setitem__(self, key: str, value: str) -> None:
        # noinspection PyTypeChecker
        key_bytes = str_to_ucs2(key)
        cdef unsigned char* key_ptr = key_bytes
        # noinspection PyTypeChecker
        value_bytes = str_to_ucs2(value)
        cdef unsigned char* value_ptr = value_bytes
        SoMR_WorkingData_SetStr(self._handle, <char16_t*>key_ptr, <char16_t*>value_ptr)

    def get_bool(self, item: str) -> bool:
        s = self[item]
        if s == "yes":
            return True
        if s == "no":
            return False
        raise TypeError(f"{item} is {s}, which is not a boolean")

    def get_int(self, item: str) -> int:
        return int(self[item])


@cython.auto_pickle(False)
cdef class OW:
    _settings: OWSettings
    _generator: OWGenerator
    _context: Context | None
    _seed: str

    @staticmethod
    def get_all_items() -> ItemList:
        # noinspection PyTypeChecker
        return ItemList.from_handle(SoMR_OW_GetAllItems())

    @staticmethod
    def get_all_locations() -> LocationList:
        # noinspection PyTypeChecker
        return LocationList.from_handle(SoMR_OW_GetAllLocations())

    @staticmethod
    cdef OWSettings _new_settings():
        # noinspection PyTypeChecker
        return OWSettings.from_handle(SoMR_OW_NewSettings())

    @staticmethod
    cdef OWGenerator _new_generator():
        # noinspection PyTypeChecker
        return OWGenerator.from_handle(SoMR_OW_NewGenerator())

    def __init__(self, src: str | pathlib.Path, seed: str, settings: dict[str, str]) -> None:
        # noinspection PyTypeChecker
        self._settings = OW._new_settings()
        for key, value in settings.items():
            self._settings[key] = value
        # noinspection PyTypeChecker
        self._generator = OW._new_generator()
        self._context = None
        self._seed = seed
        self._init(str(src))

    cdef void _init(self, src: str):
        # noinspection PyTypeChecker
        src_bytes = str_to_ucs2(src)
        cdef unsigned char* src_ptr = src_bytes
        # noinspection PyTypeChecker
        seed_bytes = str_to_ucs2(self._seed)
        cdef unsigned char* seed_ptr = seed_bytes
        handle = SoMR_OW_Init(
            <char16_t*>src_ptr,
            <char16_t*>seed_ptr,
            self._generator.get_handle(),
            self._settings.get_handle(),
        )
        # noinspection PyTypeChecker
        self._context = Context.from_handle(handle)
        error = self._context.error
        if error:
            raise Exception(error)

    def run(self, dst: str | pathlib.Path) -> None:
        # noinspection PyTypeChecker
        dst_bytes = str_to_ucs2(str(dst))
        cdef unsigned char* dst_ptr = dst_bytes
        # noinspection PyTypeChecker
        seed_bytes = str_to_ucs2(self.seed)
        cdef unsigned char* seed_ptr = seed_bytes
        SoMR_OW_Run(
            <char16_t*>dst_ptr,
            <char16_t*>seed_ptr,
            self._generator.get_handle(),
            self._settings.get_handle(),
            self._context.get_handle(),
        )
        error = self._context.error
        if error:
            raise Exception(error)

    @property
    def seed(self) -> str:
        return self._seed

    @property
    def generator(self) -> OWGenerator:
        return self._generator

    @property
    def settings(self) -> OWSettings:
        return self._settings

    @property
    def context(self) -> Context | None:
        return self._context



#!/bin/env python3

# pysomr - Python Wrapper for SoMR API
# Copyright (C) 2026  black-sliver
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# Still using setup.py (not pure pyproject) because it appears pyproject can't
# - have different compiler options per platform
# - package an extra file (SoMRandomizer.api.*) into the root of the wheel

import os
import os.path
import platform
import shutil
import subprocess  # noqa: S404: no other way to run dotnet publish
import sys
import typing as t
from pathlib import Path

from setuptools import Command, setup
from setuptools.command.build import build
from setuptools.command.build_ext import build_ext

__all__ = ["CustomCommand", "get_dotnet_os", "get_dotnet_arch"]


is_macos = platform.system() == "Darwin"
is_windows = platform.system() == "Windows"
dll_ext = ".dll" if is_windows else ".dylib" if is_macos else ".so"

debug_c_args = {
    "unix": ["-Og", "-g"],
    "gcc": ["-Og", "-g"],
}

release_c_args = {
    "unix": ["-Os", "-s", "-ffunction-sections"],
    "gcc": ["-Os", "-s", "-ffunction-sections"],
    "msvc": ["/Os"],
    "mingw32": ["-Os", "-s"],
}

debug_l_args: t.Dict[str, t.List[str]] = {}

release_l_args = {
    "unix": ["-s", "-Wl,--gc-sections", "-Wl,-rpath,$ORIGIN"],
    "gcc": ["-s", "-Wl,--gc-sections", "-Wl,-rpath,$ORIGIN"],
}

release_libraries = {
    "msvc": ["SoMRandomizer.api"],
    "unix": [":SoMRandomizer.api.so"],
    "gcc": [":SoMRandomizer.api.so"],
}

c_args = release_c_args
l_args = release_l_args
libraries = release_libraries
library_dirs = ["."]

if is_macos:
    # macOS (clang) uses some different compiler switches
    c_args_replacements: t.List[t.Tuple[str, t.Union[str, None]]] = [
        ("-s", None),
    ]
    l_args_replacements: t.List[t.Tuple[str, t.Union[str, None]]] = [
        ("-Wl,--gc-sections", None),
        ("-s", None),
        ("-Wl,-rpath,$ORIGIN", "-Wl,-rpath,@loader_path"),
    ]
    for tool in c_args:
        for orig, replacement in c_args_replacements:
            if orig in c_args[tool]:
                c_args[tool].remove(orig)
                if replacement:
                    c_args[tool].append(replacement)
    for tool in l_args:
        for orig, replacement in l_args_replacements:
            if orig in l_args[tool]:
                l_args[tool].remove(orig)
                if replacement:
                    l_args[tool].append(replacement)
    for tool in libraries:
        for i, lib in enumerate(libraries[tool]):
            # For some reason using a direct path doesn't work during link-time on macOS.
            # See explanation in CustomCommand.run.
            if lib.startswith(":SoMRandomizer.api."):
                libraries[tool][i] = "SoMRandomizer-api"
            elif lib.endswith(".so"):
                libraries[tool][i] = lib[:-3] + ".dylib"


def get_dotnet_os() -> str:
    system = platform.system()
    if system == "Windows":
        return "win"
    if system == "Darwin":
        return "osx"
    if system == "Linux":
        libc_ver = platform.libc_ver()
        if "musl" in libc_ver:
            return "linux-musl"
        if "bionic" in libc_ver:
            return "linux-bionic"
        return "linux"
    # TODO: android, ios
    return "unix"


def get_dotnet_arch() -> str:
    # FIXME: there has to be a better way?
    machine = platform.machine().lower()
    if machine in ("x86", "i386", "i586", "i686"):
        return "x86"
    if machine in ("amd64", "x86_64", "x64"):
        if sys.maxsize > 2**32:
            return "x64"
        return "x86"
    if machine in ("arm64", "aarch64"):
        return "arm64"
    if machine.startswith("arm"):
        return "arm"
    raise Exception("Unsupported architecture")


class CustomCommand(Command):
    """setuptools command to build and copy the native (AOT) lib to the lib and dist dir"""

    lib_dir: t.Union[Path, None]
    bdist_dir: t.Union[Path, None]

    def initialize_options(self) -> None:
        self.lib_dir = None
        self.bdist_dir = None

    def finalize_options(self) -> None:
        build_step = self.get_finalized_command("build_ext")
        outputs = build_step.get_outputs()
        if outputs:
            self.lib_dir = Path(outputs[0]).parent
        bdist = self.get_finalized_command("bdist_wheel")
        if bdist.bdist_dir is not None:
            self.bdist_dir = Path(bdist.bdist_dir)

    @staticmethod
    def dotnet_publish(rid: t.Optional[str] = None) -> Path:
        if rid is None:
            rid = "-".join((get_dotnet_os(), get_dotnet_arch()))
        if not rid.replace("-", "").isalnum():
            raise ValueError("Invalid rid")

        # contrary to documentation, it appears that we have to explicitly set the RID even for "native"
        # also supplying -r <rid> fails for whatever reason
        prop_dict = {
            "RuntimeIdentifier": rid,
            "OptimizationPreference": "Size",
            "InvariantGlobalization": "true",
            "UseSizeOptimizedLinq": "true",
        }
        with open("SecretOfManaRandomizer/SoMRandomizer.api/SoMRandomizer.api.csproj.user", "w") as f:
            props = "".join(f"<{key}>{value}</{key}>" for key, value in prop_dict.items())
            f.write(f"<Project><PropertyGroup>{props}</PropertyGroup></Project>")
        dotnet = shutil.which("dotnet")
        native_lib_src_name = f"SoMRandomizer.api{dll_ext}"
        output = Path(f"SecretOfManaRandomizer/SoMRandomizer.api/bin/Release/net10.0/native/{native_lib_src_name}")
        if os.environ.get("CIBUILDWHEEL"):
            # expect prebuilding the AOT lib for cibw
            print("Using prebuilt native lib")
            return output
        assert dotnet, "dotnet not found"
        subprocess.run(  # noqa: S603: no other way to run dotnet publish
            [dotnet, "publish", "SecretOfManaRandomizer/SoMRandomizer.api/"],
            check=True,
        )
        return output

    @classmethod
    def dotnet_publish_universal2(cls) -> Path:
        macos_build_dir = Path("build") / "macos"
        macos_build_dir.mkdir(parents=True, exist_ok=True)
        dotnet_os = get_dotnet_os()
        # build for arm or get prebuilt name
        arm64_src = cls.dotnet_publish(f"{dotnet_os}-arm64")
        arm64_lib = macos_build_dir / Path(arm64_src).name.replace(".dylib", ".arm64.dylib")
        # construct final name
        universal_lib = macos_build_dir / Path(arm64_src).name
        if os.environ.get("CIBUILDWHEEL"):
            # expect prebuilding the AOT lib for cibw
            print("Using prebuilt universal2 lib")
            return universal_lib
        # move arm build to temp folder
        shutil.move(arm64_src, arm64_lib)
        # build for intel
        x64_src = cls.dotnet_publish(f"{dotnet_os}-x64")
        x64_lib = macos_build_dir / Path(arm64_src).name.replace(".dylib", ".x64.dylib")
        # move intel build to temp folder
        shutil.move(x64_src, x64_lib)
        # combine the two builds into universal2
        lipo = shutil.which("lipo")
        assert lipo, "lipo not found"
        subprocess.run(  # noqa: S603: no other way to run dotnet publish
            [lipo, "-output", str(universal_lib), "-create", str(arm64_lib), str(x64_lib)],
            check=True,
        )
        # return final name
        return universal_lib

    def run(self) -> None:
        # NOTE: we need the DLL here so we can properly link it, also we want to package it further down
        if is_macos:
            native_lib_src = self.dotnet_publish_universal2()
        else:
            native_lib_src = self.dotnet_publish()
        native_lib = native_lib_src.name
        print(f"copying {native_lib_src} -> {native_lib} ({os.path.getsize(native_lib_src)})")
        shutil.copy(native_lib_src, native_lib)

        if is_macos:
            # for whatever reason it looks up the correct name during runtime, but can't link it during link time,
            # so we copy the badly-named dll to the name specified above and then ship the badly-named one anyway.
            temp_lib = f"lib{native_lib.replace('.api', '-api')}"
            print(f"copying {native_lib} -> {temp_lib}")
            shutil.copy(native_lib, temp_lib)

        if dll_ext == ".dll":
            # also copy importlib on windows
            import_lib_src = native_lib_src.with_suffix(".lib")
            print(f"copying {import_lib_src} -> {os.path.curdir} ({os.path.getsize(import_lib_src)})")
            shutil.copy(import_lib_src, ".")

        if self.lib_dir:
            print(f"copying {native_lib} -> {self.lib_dir}")
            self.lib_dir.mkdir(parents=True, exist_ok=True)
            shutil.copy(native_lib, self.lib_dir)

        if self.bdist_dir:
            print(f"copying {native_lib} -> {self.bdist_dir}")
            self.bdist_dir.mkdir(parents=True, exist_ok=True)
            shutil.copy(native_lib, self.bdist_dir)


class CustomBuildExt(build_ext):
    """customized build ext that does per-platform compiler flags"""

    def run(self) -> None:
        print("Running custom build extensions")
        super().run()

    def build_extensions(self) -> None:
        c = self.compiler.compiler_type
        if c not in c_args and c not in l_args:
            print("using unknown compiler: " + c)
        if c in c_args:
            for e in self.extensions:
                e.extra_compile_args = c_args[c]
        if c in l_args:
            for e in self.extensions:
                e.extra_link_args = l_args[c]
        if c in libraries:
            for e in self.extensions:
                e.libraries = libraries[c]
                e.library_dirs = library_dirs
        return build_ext.build_extensions(self)


class CustomBuild(build):
    sub_commands = [("build_custom", None)] + build.sub_commands


if __name__ == "__main__":
    setup(
        cmdclass={
            "build": CustomBuild,
            "build_ext": CustomBuildExt,
            "build_custom": CustomCommand,
        }
    )



from pathlib import Path
from typing import Dict, Iterator, Union

OW_SETTINGS_KEY_BOY_CLASS: str
OW_SETTINGS_KEY_GIRL_CLASS: str
OW_SETTINGS_KEY_SPRITE_CLASS: str
WORKING_DATA_KEY_BOY_CLASS: str
WORKING_DATA_KEY_BOY_EXISTS: str
WORKING_DATA_KEY_BOY_IN_LOGIC: str
WORKING_DATA_KEY_BOY_START_WEAPON_INDEX: str
WORKING_DATA_KEY_GIRL_CLASS: str
WORKING_DATA_KEY_GIRL_EXISTS: str
WORKING_DATA_KEY_GIRL_IN_LOGIC: str
WORKING_DATA_KEY_GIRL_START_WEAPON_INDEX: str
WORKING_DATA_KEY_SPRITE_CLASS: str
WORKING_DATA_KEY_SPRITE_EXISTS: str
WORKING_DATA_KEY_SPRITE_IN_LOGIC: str
WORKING_DATA_KEY_SPRITE_START_WEAPON_INDEX: str

class Context:
    error: Union[str, None]
    working_data: WorkingData

class Item:
    event_flag: int
    id: int
    name: str
    type: str

class ItemList:
    def __iter__(self) -> Iterator[Item]: ...
    def __len__(self) -> int: ...

class Location:
    event_index: int
    event_num: int
    id: int
    map_num: int
    name: str
    obj_num: int
    requirements: StrList

class LocationList:
    def __iter__(self) -> Iterator[Location]: ...
    def __len__(self) -> int: ...

class OW:
    context: Context
    generator: OWGenerator
    seed: str
    settings: OWSettings

    def __init__(self, src: Union[str, Path], seed: str, settings: Dict[str, str]) -> None: ...
    @staticmethod
    def get_all_items() -> ItemList: ...
    @staticmethod
    def get_all_locations() -> LocationList: ...
    def run(self, src: Union[str, Path]) -> None: ...

class OWGenerator:
    def get_items(self) -> ItemList: ...
    def get_locations(self) -> LocationList: ...

class OWSettings:
    def __setitem__(self, key: str, value: str) -> None: ...

class StrList:
    def __iter__(self) -> Iterator[str]: ...
    def __len__(self) -> int: ...

class WorkingData:
    def get_bool(self, key: str) -> bool: ...
    def get_int(self, key: str) -> int: ...
    def __getitem__(self, key: str) -> str: ...
